/* acertm.c -- Experimental driver for Acer Travelmate
 */

#include <linux/version.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define KERNEL26
#include <linux/moduleparam.h>
#else
#error "This module requires kernel 2.6 or later"
#endif

#if !CONFIG_PROC_FS
#error "This module requires /proc filesystem support"
#endif

#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/tty.h>
#include <linux/init.h>
#include <linux/ctype.h>
#include <linux/proc_fs.h>

/**** Module info ****/
#define DEVICE_NAME "acertm"    /* The name as displayed in /proc/modules */
#define ACERTM_VERSION "0.1"
#define ACERTM_DESCRIPTION "Acer TravelMate driver v" ACERTM_VERSION
#define ACERTM_AUTHOR "Johan Vromans <jvromans@squirrel.nl>"

MODULE_DESCRIPTION(ACERTM_DESCRIPTION);
MODULE_AUTHOR(ACERTM_AUTHOR);
MODULE_LICENSE("GPL");

/**** Module registration and release ****/
static int acertm_init(void);
static void acertm_cleanup(void);

module_init(acertm_init);
module_exit(acertm_cleanup);

/**** Driver magic constants ****/
#define ROUTINE_ARG1     ((u16)0x9610)
#define LED_NAME	 "led"	/* should be "mailled" */
#define SERVICE_LED_ON   0x131
#define SERVICE_LED_OFF  0x031
#define MARKER           0x30552142 /* "B!U0" with 16-byte alignment precedes
                             the address of bios routine */
#define ADDR_RANGE_LO    ((void *)0xF0000)
#define ADDR_RANGE_HI    ((void *)0xFFFFF)
#define ADDR_RANGE_MASK  0x0FFFF

static void *marker_addr = 0;
static void *routine_addr = 0;
static int led_init = -1;

static int led_state=0;

static void *orig_io_addr = 0;
static void *remap_io_addr = 0;
static int remap_io_size = 0;

/**** module probe/utility function prototypes ****/
static void *remap_addr(void *);
static void unmap_addr(void);
static void *find_marker(void);
static void *probe_routine_addr(void);
static u8 bios_routine(u16 service);

/* Get address for p in a block remapped from orig_io_addr to remap_io_addr */
static void *remap_addr(void *p) {
  if ( p == 0 )
    return 0;
  if ( !(orig_io_addr <= p && p < orig_io_addr + remap_io_size) ) {
    /* needs remapping */
    if (orig_io_addr != 0) {
      iounmap(remap_io_addr);
    }
    remap_io_size = ADDR_RANGE_MASK + 1;
    orig_io_addr = (void *)((u32)p & (~ADDR_RANGE_MASK));
    remap_io_addr = ioremap((int)orig_io_addr, remap_io_size);  /* TODO: error checking */
  }
  return remap_io_addr + (p - orig_io_addr);
}

static void *orig_addr(void *p) {
  if ( !remap_io_addr )
    return p;
  if ( !(remap_io_addr <= p && p < remap_io_addr + remap_io_size) )
    return p;
  return orig_io_addr + (p - remap_io_addr);
}

static void unmap_addr() {
  if ( remap_io_addr ) iounmap(remap_io_addr);
}

/**** Module probing functions ****/
/* MARKER is 16-byte aligned; the routine address is NOT 4-byte aligned;
  it occupies bytes at offset 5..7 (or 8?) from marker in little-endian order */
#include <asm/unaligned.h>
static void *probe_routine_addr() {
  u8 *p;
  if ( routine_addr ) {
    printk(DEVICE_NAME ": using address %p provided for routine\n", routine_addr);
    return routine_addr;
  }
  if ( !find_marker() )
    return 0;
  p = remap_addr(marker_addr);
  if ( !p )
    return 0;
  routine_addr = (void *)get_unaligned((u32 *)(p+5));
  printk(DEVICE_NAME ": bios routine address calculated as %p\n", routine_addr);
  return routine_addr;  /* TODO: Check range 0xE0000 - 0xFFFFF ? */
}

static void *find_marker(void) {
  u32 *p;

  if ( marker_addr != 0 ) {
    if ( marker_addr < ADDR_RANGE_LO || marker_addr > ADDR_RANGE_HI ) {
      printk(DEVICE_NAME ": marker address %p outside range %p-%p\n",
	     marker_addr, ADDR_RANGE_LO, ADDR_RANGE_HI);
      return 0;
    }
    else {
      return marker_addr;
    }
  }
  
  if ( !remap_addr(ADDR_RANGE_LO) )
    return 0;
  /* Search on 16-byte alignment */
  for ( p = remap_io_addr ; p < (u32 *)(remap_io_addr + remap_io_size); p+= 4 ) {
    if ( *p == MARKER ) {
      marker_addr = orig_addr(p);
      break;
    }
  }
  printk(DEVICE_NAME ": marker found at address %p\n", marker_addr);
  return marker_addr;
}

/**** Call to bios routine ****/
/*  Input:  ax=ROUTINE_ARG1, bx=service
    Output: al=result                      */
static u8 bios_routine(u16 service) {
  u8 result;
  void *p = remap_addr(routine_addr);

  __asm__ __volatile__ ("\n    call *%3\n"
  : "=a"(result)
  : "a"(ROUTINE_ARG1), "b"(service), "c"(p)
  );

  return result;
}

/**** Module Utility Functions ****/
static void set_led_state(int x) {
  led_state = x;
  if ( x > 0 )
    bios_routine(SERVICE_LED_ON);
  else
    if ( x == 0 )
      bios_routine(SERVICE_LED_OFF);
}

static int get_led_state(void) {
  return led_state;
}

static char* get_led_state_string(void) {
  if ( led_state > 0 )
    return "on";
  if ( led_state == 0 )
    return "off";
  return "unset";
}


/**** probe some more fun stuff ****/
static char *acertm_bios_addr = 0;

/* needle is \0 terminated string, but haystack is arbitrary of length len */
static char *find_str(char *haystack, int len, char *needle) {
  char *p;
  int nlen = strlen(needle);
 
  /* do something simple, forget automata */
  for ( p = haystack; p < haystack+len-nlen; p++ ) {
    if ( !strncmp(p, needle, nlen) )
      return p;
  }
  return 0;
}

static void probe_bios(void) {
  char *p = remap_addr(ADDR_RANGE_LO);
 
  p = find_str(p, remap_io_size, "BIOS");
  if ( p ) {
    do {
      --p;
    } while (isalnum(*p) && isascii(*p));
    p++;
  }
  acertm_bios_addr  = orig_addr(p);
}

/**** /proc interface for driver ****/
static struct proc_dir_entry *acertm_proc_root = 0;
static struct proc_dir_entry *acertm_proc_led = 0;

int acertm_proc_led_read(char *buf, char **start, off_t offset,
			 int count, int *eof, void *data) {
  *eof = 1;
  return snprintf(buf, count, "%s\n", get_led_state_string());
}

/* The last '0' or '1' in write buffer determines action taken */
int acertm_proc_led_write(struct file *file, const char *buf,
			  unsigned long count, void *data) {
  int i, state = -1;

  for ( i=0; i<count; i++ ) {
    switch ( buf[i] ) {
    case '0': state = 0; break;
    case '1': state = 1; break;
    default: break;
    }
  }
  if ( state >= 0 )
    set_led_state(state);
  return count;
}

int acertm_proc_bios_read(char *buf, char **start, off_t offset,
			  int count, int *eof, void *data) {
  char *p;
  int size;

  p = remap_io_addr;
  if ( !p ) return 0;
  size = remap_io_size - offset;
  if ( size > count ) {
    size = count;
    *eof = 0;
  }
  else {
    *eof = 1;
  }
  memcpy(buf, p + offset, size);
  *start = buf;
  return size;
}

/**** module init/cleanup code ****/
static int acertm_init() {
  printk("%s\n%s\n", ACERTM_DESCRIPTION, ACERTM_AUTHOR);

  /* Search for the bios routine */
  if ( !probe_routine_addr() ) {
    printk(DEVICE_NAME ": Bios routine not found; driver will not work.\n");
    printk(DEVICE_NAME ": Please report to authors.\n");
    return -ENODEV;
  }

  set_led_state(led_init);
  probe_bios();

  acertm_proc_root = create_proc_entry(DEVICE_NAME, S_IFDIR, 0);
  acertm_proc_led = create_proc_entry(LED_NAME, 0666, acertm_proc_root);
  acertm_proc_led->read_proc = acertm_proc_led_read;
  acertm_proc_led->write_proc = acertm_proc_led_write;

  return 0;
}

static void acertm_cleanup() {
  remove_proc_entry(LED_NAME, acertm_proc_root);
  remove_proc_entry(DEVICE_NAME, 0);
  unmap_addr();
}
